#include <stdio.h>

/*********
 * reference sequential code for the N Queen problem, SPD course
 * 2013-2014, Computer Science and Networking, University of Pisa.
 *
 * The code is a derivative work of material from wikibooks.org, and
 * as such the Creative Commons Attribution-ShareAlike License applies
 * to this code. M. Coppola
 */


/** global definitions ******************************************************/
 
// largest board size
#define MAXBOARDSIZE 24

/*********
 * levels we expand the problem of before calling the solver
 * (i.e. before spawning parallel work!)
 */
#define EXPANDLEVELS 3

/********
 * this type can be used if we need to reliably copy information about
 * a set of subproblem across parallel instances, regardless of the
 * program parameters; you will likely need to define functions that
 * copy/transfer this kind of structures across processes, threads or
 * larger arrays in order to match your parallelizatin framework of
 * choice
*/
typedef struct boardplus{
  int size; // board size
  int x; // this is where we need to restart
  int y; // idem
  int board[MAXBOARDSIZE]; // the actual board, padded to largest instance
} subproblem;

/********
 * static variable defining the board size; you may want to make it an
 * explicit parameter in your code like it's done in function main()
*/
static int board_size = 8;

/********
 * static variable counting solutions found; in a parallel settings,
 * you will need to devise a different solution
*/
static int solution_count = 0;

/* count the subproblems spawned from the expander function */
static int spawned_tasks=0;

/** utility functions ******************************************************/

/********
 * human-readable output; you may need to rewrite it in order to make
 * it really flexible */
void putboard(int rows[board_size])  
{
    int x, y;
    printf("\nSolution #%d:\n------------------------------------------------------------------\n", ++solution_count);
    for (y=0; y < board_size; ++y) {
        for (x=0; x < board_size; ++x)
            printf(x == rows[y] ? "| Q " : "|   ");
        printf("|\n------------------------------------------------------------------\n");
    }
}

// simpler printout suitable for automatic analysis
void dump_board(int rows[board_size])
{
  int y;
  printf("%d", ++solution_count);
  for (y=0; y<board_size; ++y)
    printf("\t%d",rows[y]);
  printf("\n");
}

/** computing functions ******************************************************/

/*******
 * check that a board position is safe so far; we only need to check
 * columns and diagonals, as rows are mutually safe by construction */
int is_safe(int rows[board_size], int x, int y)  
{
    int i;
    if (y == 0)
            return 1;
    for (i=0; i < y; ++i) {
       if (rows[i] == x || rows[i] == x + y - i || rows[i] == x - y +i)
            return 0;
    } 
    return 1;
}

/******* this routine fully expands a sub-instance of the problem
 * depth first, generating the corresponding set of solutions; it CAN
 * be used on the whole problem directly, if it is small enough to be
 * dealt with sequentially */
void n_queens_solver(int rows[board_size], int y) 
{
    int x;
 
    for (x=0; x < board_size; ++x) {
        if (is_safe(rows, x, y)) {
            rows[y] = x;
            if (y == board_size-1) {
	      //            putboard(rows);
	      dump_board(rows);
	    }
            else
              n_queens_solver(rows, y+1);
        }
    }
}
 
/******** this routine partially expands the problem by running the
 * combinatorial search only for the first levels; everything deeper
 * than that is passed to the solver routine, which is actually where
 * you can plug in some parallelism 
*/
void n_queens_expander(int rows[board_size], int y, int expand_levels)
{
  int x;
  
  for (x=0; x < board_size; ++x) {
    if (is_safe(rows, x, y)) // early check pruning dead branches
      {
	rows[y] = x;
	if (y == expand_levels-1) {
	  printf("New subtask %d\n", ++spawned_tasks);
	  /*******
	   * here you can spawn parallel work by filling in a
	   * supbproblem structure wit the available rows, x, y and
	   * board_size
	   */
	  n_queens_solver(rows, y+1);
	}
	else
	  n_queens_expander(rows, y+1, expand_levels);
      }
  }
}

/** main program ******************************************************/
/*******
 * Note: this very simple version of the algorithm always prints all
 * solutions. You will need to add checks to make this optional */
int main(int argc, char **argv)
{
  int rows[board_size]; // the main board variable
  int temp;
  
  if (argc ==2) {
    if (temp=atoi(argv[1]), temp >EXPANDLEVELS && temp <=MAXBOARDSIZE)
      board_size = temp;
      else {
	printf("Wrong Size parameter: %d\n",temp); 
	return -1;
      }
  }

  //  n_queens_solver(rows, 0);
  /******
   *  expand three levels, then call the full solver routine; note
   *  that this value is one of the parameters to be tuned when
   *  parallelizing 
   */
  n_queens_expander(rows, 0, 3); 

  printf("Found %d solutions over % d taks/subproblems\n",solution_count,spawned_tasks);
  return 0;
}
